@@ -3,9 +3,9 @@ use crate::instance::{InstanceData, InstancePre};
 use crate::store::StoreOpaque;
 use crate::{
     AsContextMut, Caller, Engine, Extern, ExternType, Func, FuncType, ImportType, Instance,
-    IntoFunc, Module, StoreContextMut, Trap, Val,
+    IntoFunc, Module, StoreContextMut, Trap, Val, ValRaw,
 };
-use anyhow::{anyhow, bail, Context, Error, Result};
+use anyhow::{anyhow, bail, Context, Result};
 use log::warn;
 use std::collections::hash_map::{Entry, HashMap};
 #[cfg(feature = "async")]
@@ -315,6 +315,24 @@ impl<T> Linker<T> {
         Ok(self)
     }
 
+    /// Creates a [`Func::new_unchecked`]-style function named in this linker.
+    ///
+    /// For more information see [`Linker::func_wrap`].
+    #[cfg(compiler)]
+    #[cfg_attr(nightlydoc, doc(cfg(feature = "cranelift")))] // see build.rs
+    pub unsafe fn func_new_unchecked(
+        &mut self,
+        module: &str,
+        name: &str,
+        ty: FuncType,
+        func: impl Fn(Caller<'_, T>, *mut ValRaw) -> Result<(), Trap> + Send + Sync + 'static,
+    ) -> Result<&mut Self> {
+        let func = HostFunc::new_unchecked(&self.engine, ty, func);
+        let key = self.import_key(module, Some(name));
+        self.insert(key, Definition::HostFunc(Arc::new(func)))?;
+        Ok(self)
+    }
+
     /// Creates a [`Func::new_async`]-style function named in this linker.
     ///
     /// For more information see [`Linker::func_wrap`].
@@ -645,21 +663,14 @@ impl<T> Linker<T> {
                                 // `unwrap()` everything here because we know the instance contains a
                                 // function export with the given name and signature because we're
                                 // iterating over the module it was instantiated from.
-                                let command_results = instance
+                                instance
                                     .get_export(&mut caller, &export_name)
                                     .unwrap()
                                     .into_func()
                                     .unwrap()
-                                    .call(&mut caller, params)
+                                    .call(&mut caller, params, results)
                                     .map_err(|error| error.downcast::<Trap>().unwrap())?;
 
-                                // Copy the return values into the output slice.
-                                for (result, command_result) in
-                                    results.iter_mut().zip(command_results.into_vec())
-                                {
-                                    *result = command_result;
-                                }
-
                                 Ok(())
                             },
                         )
@@ -718,20 +729,14 @@ impl<T> Linker<T> {
                                 let (instance_pre, export_name) = &*upvars;
                                 let instance = instance_pre.instantiate_async(&mut caller).await?;
 
-                                let command_results = instance
+                                instance
                                     .get_export(&mut caller, &export_name)
                                     .unwrap()
                                     .into_func()
                                     .unwrap()
-                                    .call_async(&mut caller, params)
+                                    .call_async(&mut caller, params, results)
                                     .await
                                     .map_err(|error| error.downcast::<Trap>().unwrap())?;
-
-                                for (result, command_result) in
-                                    results.iter_mut().zip(command_results.into_vec())
-                                {
-                                    *result = command_result;
-                                }
                                 Ok(())
                             })
                         },
@@ -1035,22 +1040,11 @@ impl<T> Linker<T> {
         let store = store.as_context_mut().0;
         let imports = module
             .imports()
-            .map(|import| {
-                self._get_by_import(&import)
-                    .ok_or_else(|| self.link_error(&import))
-            })
+            .map(|import| self._get_by_import(&import))
             .collect::<Result<_>>()?;
         unsafe { InstancePre::new(store, module, imports) }
     }
 
-    fn link_error(&self, import: &ImportType) -> Error {
-        let desc = match import.name() {
-            Some(name) => format!("{}::{}", import.module(), name),
-            None => import.module().to_string(),
-        };
-        anyhow!("unknown import: `{}` has not been defined", desc)
-    }
-
     /// Returns an iterator over all items defined in this `Linker`, in
     /// arbitrary order.
     ///
@@ -1128,16 +1122,20 @@ impl<T> Linker<T> {
     ) -> Option<Extern> {
         let store = store.as_context_mut().0;
         // Should be safe since `T` is connecting the linker and store
-        Some(unsafe { self._get_by_import(import)?.to_extern(store) })
+        Some(unsafe { self._get_by_import(import).ok()?.to_extern(store) })
     }
 
-    fn _get_by_import(&self, import: &ImportType) -> Option<Definition> {
-        if let Some(item) = self._get(import.module(), import.name()) {
-            return Some(item.clone());
+    fn _get_by_import(&self, import: &ImportType) -> anyhow::Result<Definition> {
+        fn undef_err(missing_import: &str) -> anyhow::Error {
+            anyhow!("unknown import: `{}` has not been defined", missing_import)
         }
 
-        if import.name().is_some() {
-            return None;
+        if let Some(item) = self._get(import.module(), import.name()) {
+            return Ok(item.clone());
+        }
+
+        if let Some(name) = import.name() {
+            return Err(undef_err(&format!("{}::{}", import.module(), name)));
         }
 
         if let ExternType::Instance(t) = import.ty() {
@@ -1158,13 +1156,15 @@ impl<T> Linker<T> {
             // suffice.
             let mut map = indexmap::IndexMap::new();
             for export in t.exports() {
-                let item = self._get(import.module(), Some(export.name()))?;
+                let item = self
+                    ._get(import.module(), Some(export.name()))
+                    .ok_or_else(|| undef_err(&format!("{}::{}", import.module(), export.name())))?;
                 map.insert(export.name().to_string(), item.clone());
             }
-            return Some(Definition::Instance(Arc::new(map)));
+            return Ok(Definition::Instance(Arc::new(map)));
         }
 
-        None
+        Err(undef_err(&import.module()))
     }
 
     /// Returns the "default export" of a module.
